## Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
## SPDX-License-Identifier: MIT-0

import subprocess
import sys

# Install missing dependencies during script execution
def install(package):
    subprocess.check_call([sys.executable, "-m", "pip", "install", package])


install('wfdb')
install('sagemaker')

import boto3
import sagemaker
from sagemaker import get_execution_role
from sagemaker.sklearn.processing import SKLearnProcessor
import os
import re
import numpy as np
import json
import wfdb
from joblib import Parallel, delayed
import multiprocessing
import pandas as pd

# Convert command line args into a map of args
args_iter = iter(sys.argv[1:])
args = dict(zip(args_iter, args_iter))

input_dir = args['input_path']
output_dir = args['output_path']
group_data_out = args['group_data_out']

try:
    os.mkdir(output_dir)
except FileExistsError:
    pass


def get_wav_vals(wav_dat):
    """
    Get waveform statistics and metadata.
    
    This function gets waveform summary statistics data for a dat file.
    
    :param wav_dat: wfdb data record object
    :type wav_dat: wfdb record
    :return: dictionary of record statistics and metadata
    :rtype: dict
    """
    wav = wav_dat.__dict__
    sig_dat = wav['p_signal']
    sig_name = wav['sig_name']
    record_name = wav['record_name']
    wav_summary = get_summary(sig_dat, sig_name, record_name)

    return wav_summary


def compute_waveform_stats(sig_dat):
    """
    Compute waveform statistics.
    
    :param sig_dat: list of lists containing signal data
    :type sig_dat: list
    :return: tuple of floats
    :rtype: tuple
    """
    summ = np.nansum(sig_dat, axis=0)
    mean = np.nanmean(sig_dat, axis=0)
    sigma = np.nanstd(sig_dat, axis=0)
    signal_length = sum(~np.isnan(sig_dat))  # Account for NaNs
    q5 = np.nanquantile(sig_dat, q=0.05, axis=0)
    q95 = np.nanquantile(sig_dat, q=0.95, axis=0)
    q10 = np.nanquantile(sig_dat, q=0.1, axis=0)
    q97p5 = np.nanquantile(sig_dat, q=0.975, axis=0)
    minn = np.nanmin(sig_dat, axis=0)
    maxx = np.nanmax(sig_dat, axis=0)
    med = np.nanmedian(sig_dat, axis=0)    
    variance = np.nanvar(sig_dat, axis=0)
    ssq = np.nansum(sig_dat**2, axis=0)
    
    return summ, mean, sigma, signal_length, q5, q95, variance, ssq, q10, q97p5, minn, maxx, med


def get_summary(sig_dat, sig_name, record_name):
    """
    Compiles waveform signal statistics for a dat file.
    
    This function identifies the waveform type for a dat file, gets the dat file's signal statistics,
    and then compiles them into a dictionary.
    
    :param sig_dat: list of lists containing signal data
    :param sig_name: list of signal names
    :param record_name: name of dat file's record
    :type sig_dat: list
    :type sig_name: list
    :type record_name: str
    :return: dictionary of record statistics and metadata
    :rtype: dict
    """
    record_vals = {}
    record_id = [re.split('_|n', record_name)[0] for i in range(sig_dat.shape[1])]
    waveform_id = [record_name for i in range(sig_dat.shape[1])]
    if 'n' in record_name:
        waveform_type = 'numeric'
    else:
        waveform_type = 'segment'
    waveform_type = [waveform_type for i in range(sig_dat.shape[1])]
    
    summ, mean, sigma, signal_length, q5, q95, variance, ssq, q10, q97p5, minn, maxx, med = compute_waveform_stats(sig_dat)
    sig_dat[sig_dat == 0] = np.nan
    summ_zasn, mean_zasn, sigma_zasn, signal_length_zasn, q5_zasn, q95_zasn, variance_zasn, ssq_zasn, q10_zasn, q97p5_zasn, minn_zasn, maxx_zasn, med_zasn = compute_waveform_stats(sig_dat)

    record_vals['record_id'] = dict(zip(sig_name, record_id))
    record_vals['waveform_id'] = dict(zip(sig_name, waveform_id))
    record_vals['waveform_type'] = dict(zip(sig_name, waveform_type))
    
    record_vals['sum_no_nans'] = dict(zip(sig_name, summ))
    record_vals['mean_no_nans'] = dict(zip(sig_name, mean))
    record_vals['sigma_no_nans'] = dict(zip(sig_name, sigma))
    record_vals['sum_of_squares_no_nans'] = dict(zip(sig_name, ssq))
    record_vals['signal_length_no_nans'] = dict(zip(sig_name, signal_length))
    record_vals['variance_no_nans'] = dict(zip(sig_name, variance))
    record_vals['quantile_5_no_nans'] = dict(zip(sig_name, q5))
    record_vals['quantile_95_no_nans'] = dict(zip(sig_name, q95))
    record_vals['quantile_10_no_nans'] = dict(zip(sig_name, q10))
    record_vals['quantile_97p5_no_nans'] = dict(zip(sig_name, q97p5))
    record_vals['min_no_nans'] = dict(zip(sig_name, minn))
    record_vals['max_no_nans'] = dict(zip(sig_name, maxx))
    record_vals['median_no_nans'] = dict(zip(sig_name, med))
    
    record_vals['sum_no_nans_no_zeros'] = dict(zip(sig_name, summ_zasn))
    record_vals['mean_no_nans_no_zeros'] = dict(zip(sig_name, mean_zasn))
    record_vals['sigma_no_nans_no_zeros'] = dict(zip(sig_name, sigma_zasn))
    record_vals['sum_of_squares_no_nans_no_zeros'] = dict(zip(sig_name, ssq_zasn))
    record_vals['signal_length_no_nans_no_zeros'] = dict(zip(sig_name, signal_length_zasn))
    record_vals['variance_no_nans_no_zeros'] = dict(zip(sig_name, variance_zasn))
    record_vals['quantile_5_no_nans_no_zeros'] = dict(zip(sig_name, q5_zasn))
    record_vals['quantile_95_no_nans_no_zeros'] = dict(zip(sig_name, q95_zasn))
    record_vals['quantile_10_no_nans_no_zeros'] = dict(zip(sig_name, q10_zasn))
    record_vals['quantile_97p5_no_nans_no_zeros'] = dict(zip(sig_name, q97p5_zasn))
    record_vals['min_no_nans_no_zeros'] = dict(zip(sig_name, minn_zasn))
    record_vals['max_no_nans_no_zeros'] = dict(zip(sig_name, maxx_zasn))
    record_vals['median_no_nans_no_zeros'] = dict(zip(sig_name, med_zasn))

    return record_vals


def combine_summary_data(dat, input_path, output_path):
    """
    Process dat record.
    
    This function reads a dat file's contents, gets its waveform statistics and metadata,
    and returns the dat file's waveform summary.
    
    :param dat: dat file name
    :param input_path: dat input directory path
    :param output_path: json output directory path
    :type dat: str
    :type input_path: str
    :type output_path: str
    :return: dataframe containing dat file's waveform summary data
    :rtype: pandas dataframe
    """
    try:
        wav_dat = wfdb.rdrecord(os.path.join(input_path, dat))
        wav_summary = get_wav_vals(wav_dat)
        return pd.DataFrame(wav_summary)
    except IndexError as e:
        print("WFDB IndexError when reading record", dat, ":", e)
        pass


def run_process(dat_chunk, input_path=input_dir, output_path=output_dir):
    """
    Create combined data output for a chunk of dat files.
    
    This function loops over and processes all dat files present in dat_chunk. It
    combines each dat file's processed outputs into a single dat_chunk dataframe.
    
    :param dat_chunk: list of dat file names
    :param input_path: input directory
    :param output_path: output directory
    :type dat_chunk: list
    :type input_path: str
    :type output_path: str
    :return: dataframe containing all processed data for files included in dat_chunk
    :rtype: pandas dataframe
    """
    ws_df = pd.DataFrame()
    for dat in dat_chunk:
        ws_df = ws_df.append(combine_summary_data(dat, input_path, output_path=output_path))
    return ws_df


def chunks(lst, n):
    """
    Yield successive n-sized chunks from lst.
    
    :param lst: list of elements to be divided
    :param n: number of elements per chunk
    :type lst: list
    :type n: int
    :return: generator comprising evenly sized chunks
    :rtype: class 'generator'
    """
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def main(*args, **kwarg):
    """
    Execute first pass preprocessing logic to obtain initial waveform statistics
    
    :param args: positional arguments
    :param kwargs: keyword arguments
    :return: None
    """
    # Generate list of .dat files on machine
    data_dir = input_dir
    d_subs = next(os.walk(os.path.join(data_dir, '.')))[1]
    file_list = []
    for ds in d_subs:
        file_list.extend(os.listdir(os.path.join(data_dir, ds, '.')))
    dat_list = [os.path.join(re.split('_|\.', f)[0].replace('n', ''), f[:-4]) for f in file_list if f[-4:] == '.dat']

    # Split list of files into sub-lists
    cpu_count = multiprocessing.cpu_count()
    splits = int(len(dat_list) / cpu_count)
    if splits == 0: splits = 1
    dat_chunks = list(chunks(dat_list, splits))

    # Parallelize processing of sub-lists across CPUs
    ws_df_list = Parallel(n_jobs=-1, verbose=0)(delayed(run_process)(dc) for dc in dat_chunks)
#     ws_df_list = Parallel(n_jobs=-1, verbose=0)(delayed(return_chunks)(dc) for dc in dat_chunks)

    # Compile and pickle patient group dataframe
    ws_df_group = pd.concat(ws_df_list)
    ws_df_group = ws_df_group.reset_index().rename(columns={'index': 'signal'})
    ws_df_group.to_json(os.path.join(output_dir, group_data_out))


# Check to see if this file is the "__main__" script being executed
if __name__ == '__main__':
    main()