## Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
## SPDX-License-Identifier: MIT-0

import subprocess
import sys

# Install missing dependencies during script execution
def install(package):
    subprocess.check_call([sys.executable, "-m", "pip", "install", package])

install('awswrangler')
import awswrangler as wr

import json
import boto3
import random
import numpy as np
from io import StringIO
import pandas as pd

# Convert command line args into a map of args
args_iter = iter(sys.argv[1:])
args = dict(zip(args_iter, args_iter))

bucket = args['bucket']
pass1out_prefix = args['pass1out_prefix']
pass1out_data = args['pass1out_data']
pass1p5out_data = args['pass1p5out_data']
statsdata_name = args['statsdata_name']


def normalize_signal_names(df):
    """
    Normalize signal names.
    
    :param df: dataframe containing signal data
    :type df: pandas dataframe
    :return: dataframe with normalized signal names
    :rtype: pandas dataframe
    """
    df['signal'] = df['signal'].str.upper()
    for sig in df['signal'].unique():
        if sig in ['ABPDIAS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ABP DIAS'
        elif sig in ['ABPMEAN']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ABP MEAN'
        elif sig in ['ABPSYS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ABP SYS'
        elif sig in ['ARTDIAS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ART DIAS'
        elif sig in ['ARTMEAN']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ART MEAN'
        elif sig in ['ARTSYS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ART SYS'
        elif sig in ['NBPDIAS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'NBP DIAS'
        elif sig in ['NBPMEAN']:
            df.loc[(df['signal'] == sig), 'signal'] = 'NBP MEAN'
        elif sig in ['NBPSYS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'NBP SYS'
        elif sig in ['PAPDIAS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'PAP DIAS'
        elif sig in ['PAPMEAN']:
            df.loc[(df['signal'] == sig), 'signal'] = 'PAP MEAN'
        elif sig in ['PAPSYS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'PAP SYS'
        elif sig in ['ST-I']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ST I'
        elif sig in ['ST-II']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ST II'
        elif sig in ['ST-III']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ST III'
        elif sig in ['ST-MCL']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ST MCL'
        elif sig in ['ST-V']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ST V'
        elif sig in ['ST-AVF']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ST AVF'
        elif sig in ['ST-AVL']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ST AVL'
        elif sig in ['ST-AVR']:
            df.loc[(df['signal'] == sig), 'signal'] = 'ST AVR'
        elif sig in ['%SPO2', 'SPO2 R', 'SPO2 L', 'SPO2_AP', 'SPO2 APERIODIC']:
            df.loc[(df['signal'] == sig), 'signal'] = 'SPO2'
        elif sig in ['BAPSYS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'BAP SYS'
        elif sig in ['BAPDIAS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'BAP DIAS'
        elif sig in ['BAPMEAN']:
            df.loc[(df['signal'] == sig), 'signal'] = 'BAP MEAN'
        elif sig in ['AOBPSYS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'AOBP SYS'
        elif sig in ['AOBPDIAS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'AOBP DIAS'
        elif sig in ['AOBPMEAN']:
            df.loc[(df['signal'] == sig), 'signal'] = 'AOBP MEAN'
        elif sig in ['AOSYS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'AO SYS'
        elif sig in ['AODIAS']:
            df.loc[(df['signal'] == sig), 'signal'] = 'AO DIAS'
        elif sig in ['AOMEAN']:
            df.loc[(df['signal'] == sig), 'signal'] = 'AO MEAN'

    return df


def get_data():
    """
    Get and combine patient group output data from pass 1.
    
    :return: waveform summary dataframe
    :rtype: pandas dataframe
    """
    s3 = boto3.client('s3')
    pat_groups = list(range(30, 40))
    wavs_df = pd.DataFrame()
    for group in pat_groups:
        group_json = json.loads(
            s3.get_object(Bucket=bucket, Key='/'.join([pass1out_prefix, str(group), pass1out_data]))['Body'].read())
        wavs_df = wavs_df.append(pd.DataFrame(group_json))
    print(wavs_df.signal.unique())

    return wavs_df


def write_parquet(wavs_df, path):
    """
    Write waveform summary dataframe to S3 in parquet format.
    
    :param wavs_df: waveform summary dataframe
    :param path: S3 directory prefix
    :type wavs_df: pandas dataframe
    :type path: str
    :return: None
    """
    extra_args = {"ServerSideEncryption": "aws:kms"}
    wr.s3.to_parquet(
        df=wavs_df,
        path=path,
        compression='snappy',
        s3_additional_kwargs=extra_args)


def combine_data():
    """
    Get combined data and write to parquet.
    
    :return: waveform summary dataframe
    :rtype: pandas dataframe
    """
    wavs_df = get_data()
    wavs_df = normalize_signal_names(wavs_df)
    write_parquet(wavs_df, "s3://{}/{}/{}".format(bucket, pass1out_prefix, pass1p5out_data))

    return wavs_df


def subset_columns(wavs_df):
    """
    Subset waveform summary dataframe to only columns of interest.
    
    :param wavs_df: waveform summary dataframe
    :type wavs_df: pandas dataframe
    :return: tuple of dataframes and signal names
    :rtype: tuple
    """
    no_nanzero_cols = [c for c in wavs_df.columns if 'no_nans_no_zeros' in c]
    selection = ['record_id', 'signal'] + no_nanzero_cols
    print(selection)
    cleaned_df = wavs_df[selection]
    signals = cleaned_df['signal'].unique()

    population_stats_df = pd.DataFrame(index=signals, columns=['mean',
                                                               'sigma',
                                                               'length',
                                                               'mean-2sigma',
                                                               'mean+2sigma',
                                                               'avg_5perc_quantile',
                                                               'avg_95perc_quantile',
                                                               'min_5perc_quantile',
                                                               'max_95perc_quantile',
                                                               'min_10perc_quantile',
                                                               'max_97p5perc_quantile'
                                                               ])

    return cleaned_df, signals, population_stats_df


def range_stats(df):
    """
    Calculate statistics potentially required to obtain bin and quantile ranges.
    
    :param df: signal-specific waveform summary dataframe
    :type df: pandas dataframe
    :return: tuple of statistic float values
    :rtype: tuple
    """
    sums = df[['sum_of_squares_no_nans_no_zeros',
               'sum_no_nans_no_zeros',
               'signal_length_no_nans_no_zeros',
               'quantile_5_no_nans_no_zeros',
               'quantile_95_no_nans_no_zeros',
              ]].sum()
    mean = sums[1]/sums[2]
    var = sums[0] / sums[2] - mean ** 2
    min_q5 = df['quantile_5_no_nans_no_zeros'].min()
    max_q95 = df['quantile_95_no_nans_no_zeros'].max()
    min_q10 = df['quantile_10_no_nans_no_zeros'].min()
    max_q97p5 = df['quantile_97p5_no_nans_no_zeros'].max()
    
    return mean, var, sums[2], sums[3]/df.shape[0], sums[4]/df.shape[0], min_q5, max_q95, min_q10, max_q97p5


def write_csv(df, file_name):
    """
    Write population statistics data CSV to S3
    
    :param df: population statistics dataframe
    :param file_name: name of CSV to be written
    :type df: pandas dataframe
    :type file_name: str
    :return: None
    """
    s3 = boto3.client('s3')
    csv_buffer = StringIO()
    df.to_csv(csv_buffer, index=True)
    s3.put_object(Bucket=bucket, Key='/'.join([pass1out_prefix, file_name]), Body=csv_buffer.getvalue(), ServerSideEncryption='aws:kms')


def create_stats(wavs_df):
    """
    Calculate population statistics by signal.
    
    :param wavs_df: waveform summary dataframe
    :type wavs_df: pandas dataframe
    :return: None
    """
    df_for_stats, signals, population_stats_df = subset_columns(wavs_df)
    for signal in signals:
        m, v, l, q5, q95, q5min, q95max, q10, q97p5 = range_stats(df_for_stats.loc[df_for_stats['signal'] == signal,])
        if v < 0:
            print(signal, v, m, l)
        else:
            population_stats_df.loc[signal][
                'mean', 'sigma', 'length', 'mean-2sigma', 'mean+2sigma', 'avg_5perc_quantile',
                'avg_95perc_quantile', 'min_5perc_quantile', 'max_95perc_quantile',
                'min_10perc_quantile', 'max_97p5perc_quantile'] = [m, np.sqrt(v), l, m - 2 * np.sqrt(v),
                                                                   m + 2 * np.sqrt(v), q5, q95, q5min, q95max, q10,
                                                                   q97p5]
    population_stats_df.index.name="signal"
    write_csv(population_stats_df, statsdata_name)


def main(*args, **kwarg):
    """
    Execute pass 1.5 preprocessing logic to post-process pass 1 outputs,
    generate train/val/test splits, and compute population summary statistics.
    
    :param args: positional arguments
    :param kwargs: keyword arguments
    :return: None
    """
    wavs_df = combine_data()
    create_stats(wavs_df)


# Check to see if this file is the "__main__" script being executed
if __name__ == '__main__':
    main()